# Copyright (C) 2000 Christopher Craig
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

"""Parses PyHP documents
"""

from mod_python import apache
import code
from UserString import UserString
import os
import sys
import re

__version__ = "$Id: pyhp_module.py,v 1.14 2000/11/10 17:26:29 ccraig Exp $"

try:
    # The CGI spec requires changing directory, which is not thread safe
    import threading
    _lock = threading.Lock()
except (ImportError, AttributeError):
    class DummyLock:
        def acquire(self):
            pass
        def release(self):
            pass
    _lock = DummyLock()

os.environ = {}
reg = re.compile('(<%.*?%>)', re.S)
staticvars = {}

class fbuffer(UserString):
    """Acts sort of like an mmap'ed file with out the file.
    Allows the use of a string object as sys.stdout
    """
    
    def __init__(self):
        UserString.__init__(self, '')
        
    def write(self, str):
        self.data += str


class PageData:
    """Allows the modification of body and headers on the fly"""

    def __init__(self, req):
        self._body = fbuffer()
        self._headers = {}
        self._req = req
        if req.get_options().has_key('AllowStatic'):
            self.staticvars = staticvars
        else:
            self.staticvars = {} # allows code portability
        
    def addheader(self, invar):
        """
        takes a HTTP header string (of the form 'Key: value') and adds it
        to the header section.  This can be done anywhere in the page.
        """
        s = str(invar)
        t = s.split(':', 1)
        self._headers[t[0]] = t[1]
        
    def clearall(self):
        """
        clears all data output either to the stdout or to the headers up
        to this point, starting as if no code had been executed and no text
        had been processed.  This is useful for generating error pages.
        """
        self._body.data = ''
        self._headers = {}

    def _writeout(self):
        """sends page out to port, headers first"""

        # process headers
        self._req.content_type = 'text/html'
        for k, v in self._headers.items():
            if k.lower() == 'content-type':
                self._req.content_type = v
                self._req.headers_out[k] = v
            elif k.lower() == 'status':
                self._req.status = v.split()[0]
            else:
                self._req.headers_out.add(k, v)

        # send headers to port
        self._req.send_http_header()

        # send body to port
        self._req.write(str(self._body))

    def include(self, file):
        """
        includes file (a file object) as a PyHP file.
        This takes place in a fresh namespace and does not inherit variables
        from the current namespace.  Text and headers generated by the PyHP
        file are inserted at the present place in the document body and
        headers respectively.  This procedure is at least an order of
        magnitude slower than a python 'import' statement, so should be used
        sparingly
        """

        cgienv = {'pyhp': self,
                  'sys':  sys}
        interp = PyHPInterp(cgienv)

        # read in file
        try:
            file.seek(0,0)
            data = file.read()
            if data[0] == '#': # detect shell script syntax
                data = data.split('\n', 1)[1]
        except IOError:
            raise apache.SERVER_RETURN, apache.HTTP_INTERNAL_SERVER_ERROR

        # split file into sections
        fields = reg.split(data)
        fields = filter(lambda x: len(x)!=0, fields)

        # setup data and parse code
        for f in fields:
            if f[:3] == '<%=': interp.pushvar(f[4:-2])
            elif f[:3] == '<%-': pass
            elif f[:2] == '<%': interp.pushcode(f[3:-2])
            else: interp.pushtext(f)

        
    
class PyHPInterp(code.InteractiveInterpreter):
    """Interpreter for PyHP.  Allows mod_python to spawn a seperate
    pseudo interpreter for each script run with its own namespace, but still
    cache module loads"""
    
    def pushcode(self, codeobj):
        c = code.compile_command(codeobj, '<string>', 'exec')
        if c:
            self.runcode(c)
        else:
            raise SyntaxError, 'Code segment does not compile'
                    
    def pushvar(self, var):
        cmd = 'sys.stdout.write(str(%s))' % var
        self.runsource(cmd)
        
    def pushtext(self, text):
        self.locals['sys'].stdout.write(text)
        

def handler(req):
    """handler function for mod_python"""

    if req.subprocess_env.has_key("script_filename"):
        dir, file = os.path.split(req.subprocess_env["script_filename"])
    else:
        dir, file = os.path.split(req.filename)

    _lock.acquire()

    try:
        try:
            # store old environment
            old_cwd = os.getcwd()
            env = os.environ
            so = sys.__stdout__

            # setup CGI environment
            os.chdir(dir)
            pyhp = PageData(req)
            os.environ=apache.build_cgi_env(req)
            sys.stdout = sys.__stdout__ = pyhp._body # points to buffer
            sys.stdin = apache.CGIStdin(req)   # points to request port
            sys.argv = []

            # execute code
            pyhp.include(open(file))

            # print content
            pyhp._writeout()

            return apache.OK
            
        finally:
            # reset environment
            os.environ = env
            sys.stdout = sys.__stdout__ = so
            sys.stdin = sys.__stdin__

            os.chdir(old_cwd)

            

    finally:
        _lock.release()

            
